package main;

import main.HexCell.Direction;
import main.Visualizer.Layer;

/**
 * Hangyákat reprezentáló osztály
 * 
 * @author Kozaroczy
 * 
 */
public class Ant extends Wanderer {

	/**
	 * A hangya állapotára vonatkozó enum, lehet AS_Searching (keresés),
	 * AS_Returning (visszatérés) vagy AS_Dead (halott)
	 * 
	 */
	public enum AntState {
		AS_Searching, AS_Returning, AS_Dead
	}

	/**
	 * A hangyánál lévő étel. Amikor felveszi, csak ő hivatkozik rá, a mező nem
	 * 
	 * @see #getFood()
	 * @see #setFood(Food)
	 */
	protected Food myFood;
	/**
	 * A hangya aktuális állapota
	 * 
	 * @see AntState
	 */
	protected AntState state;
	/** A hangya alapértelmezett sebessége: {@value} */
	protected static final int defaultSpeed = 2;
	/** A hangya színe */
	protected boolean isRed;

	/**
	 * Hangya konstruktora
	 * 
	 * @param location
	 *            Alapértelmezett pozíció
	 * @see Wanderer#Wanderer(HexCell)
	 * 
	 */
	Ant(HexCell location) {
		super(location);
		Logger.entering(this, new Object[] { location });
		/** Alapértékek beállítása */
		myFood = null;
		speed = defaultSpeed;
		state = AntState.AS_Searching;
		isRed = (int) Math.floor(Math.random() * 2) == 1;

		/** Szól, hogy új hangya keletkezett */
		Director.getInstance().incAnt();
		Logger.exiting(this);
	}

	/**
	 * Minden paraméterre kiterjedő konstruktor a pályabetöltéshez
	 * 
	 * @param location
	 *            A hangya helye
	 * @param dir
	 *            A hangya iránya
	 * @param hasFood
	 *            Van-e étel a hangyánál
	 * @param speed
	 *            A hangya sebessége
	 * 
	 */
	Ant(HexCell location, Direction dir, int speed, boolean hasFood) {
		super(location, dir, speed);
		Logger.entering(this, new Object[] { location, dir, hasFood, speed });
		/** Értékek beállítása */
		myFood = hasFood ? new Food(null) : null;
		state = hasFood ? AntState.AS_Returning : AntState.AS_Searching;

		/** Szól, hogy új hangya keletkezett */
		Director.getInstance().incAnt();
		Logger.exiting(this);
	}

	/**
	 * Meghívja az {@link IInterActor#interactWith(Ant)} függvényét,
	 * paraméterként adva saját magát.
	 * 
	 * @param interactor
	 *            A hívó fél
	 */
	public void accept(IInterActor interactor) {
		Logger.entering(this, new Object[] { interactor });
		interactor.interactWith(this);
		Logger.exiting(this);
	}

	/**
	 * A hangya ennek hatására {@linkplain AntState AS_Dead} (halott) állapotba
	 * kerül, és ha volt nála étel, akkor eldobja
	 * 
	 * @see AntState
	 * 
	 */
	public void die() {
		Logger.entering(this, null);
		/** Ha volt nála étel, eldobja */
		if (state == AntState.AS_Returning) {
			myFood.drop(location);
		}
		state = AntState.AS_Dead;
		Director.getInstance().decAnt();
		super.dispose();
		Logger.exiting(this);
	}

	/**
	 * {@inheritDoc}
	 */
	public String draw() {
		String suffixes = "_" + direction.toString();
		if (isRed)
			suffixes = "_red" + suffixes;
		return "img/ant" + suffixes + ".png";
	}

	/**
	 * A hangyánál lévő ételt adja vissza.
	 * 
	 * @return Hangyánál lévő étel
	 * 
	 */
	public Food getFood() {
		Logger.entering(this, null);
		Logger.exiting(this, myFood);
		return myFood;
	}

	/**
	 * {@inheritDoc}
	 */
	public Layer getLayer() {
		return Layer.ACTOR;
	}

	/**
	 * {@inheritDoc} Ha van nála étel,
	 * jelezi
	 * 
	 * @param a
	 *            Hangyaboly
	 * @see AntHill#foodCollected()
	 * @see AntState
	 */
	public void interactWith(AntHill a) {
		Logger.entering(this, new Object[] { a });
		if (state == AntState.AS_Returning) {
			a.foodCollected();
			myFood.smallImage.removeFromCell();
			setFood(null);
			state = AntState.AS_Searching;
		}
		Logger.exiting(this);
	}

	/**
	 * {@inheritDoc} Meghal. Ha volt nála étel, máshova kerül.
	 * 
	 * @see Antlion#foodDestroyed()
	 * @see #die()
	 */
	public void interactWith(Antlion a) {
		Logger.entering(this, new Object[] { a });
		if (state == AntState.AS_Returning) {
			a.foodDestroyed();
			setFood(null);
			state = AntState.AS_Searching;
		}
		if (state != AntState.AS_Dead) {
			die();
		}
		Logger.exiting(this);
	}

	/**
	 * {@inheritDoc} Visszahívja az
	 * {@link Echidna#interactWith(Ant)} függvényt
	 * 
	 * @param e
	 *            Hangyászsün
	 */
	public void interactWith(Echidna e) {
		Logger.entering(this, new Object[] { e });
		e.interactWith(this);
		Logger.exiting(this);
	}

	/**
	 * {@inheritDoc} Ha tudja, felveszi.
	 * 
	 * @param f
	 *            Étel
	 * @see AntState
	 */
	public void interactWith(Food f) {
		Logger.entering(this, new Object[] { f });
		if (state == AntState.AS_Searching) {
			location.removeObject(f);
			setFood(f);
			state = AntState.AS_Returning;
		}
		Logger.exiting(this);
	}

	/**
	 * {@inheritDoc} Ha vitt ételt, újragondolja az útvonalát.
	 */
	public void interactWith(Pebble p) {
		Logger.entering(this, new Object[] { p });
		if (state == AntState.AS_Returning)
			Director.getInstance().generatePaths();
		super.interactWith(p);
		Logger.exiting(this);
	}

	/**
	 * {@inheritDoc} Meghal tőle.
	 * 
	 * @param p
	 *            Méregfolt
	 * @see #die()
	 */
	public void interactWith(Poison p) {
		Logger.entering(this, new Object[] { p });
		if (state != AntState.AS_Dead)
			die();
		Logger.exiting(this);
	}

	/**
	 * A hangya mozgása a paraméterként kapott cellába. A
	 * {@link Wanderer#move(HexCell)} függvényt hívja tovább, majd
	 * {@linkplain AntScent szagnyomot} hagy
	 * 
	 * @param next
	 *            Következő cella
	 * 
	 */
	public void move(HexCell next) {
		Logger.entering(this, new Object[] { next });
		if (state == AntState.AS_Returning) {
			myFood.smallImage.removeFromCell();
		}
		speed = defaultSpeed;
		super.move(next);
		if (state == AntState.AS_Returning && myFood.smallImage.location == null) {
			myFood.smallImage.addToCell(location);
		}
		if (state != AntState.AS_Dead) {
			new AntScent(this.location);
		}
		Logger.exiting(this);
	}

	/**
	 * A paraméterül kapott változót elmenti a {@link #myFood} változóba.
	 * 
	 * @param newVal
	 *            Új étel
	 */
	public void setFood(Food newVal) {
		if (newVal != null) {
			Logger.entering(this, new Object[] { newVal });
		} else {
			Logger.entering(this, null);
		}
		myFood = newVal;
		Logger.exiting(this);
	}

	/**
	 * {@inheritDoc} A következő irányt a szomszédos cellák
	 * {@linkplain Located#influence() befolyása} alapján állapítja meg kereső
	 * állapotban. Visszatérő állapotban a legközelebbi hangyaboly felé mozog.
	 * 
	 */
	public void tick() {
		Logger.entering(this, null);
		unusedTicks++;
		/** Nem minden tick-ben mozog */
		if (unusedTicks >= speed) {
			if (state == AntState.AS_Searching) {
				/** Ha keresgélés közben mozog */

				/** Lekéri a szemben lévő cellát */
				HexCell straight = location.getCell(direction);
				/**
				 * Ha van ott cella, akkor lekéri az influence-t, egyébként
				 * érvénytelen adatot állít be
				 */
				int st_influence = (straight == null) ? -9 : straight.getInfluence();
				/** Alapesetben előre halad inkább, így oda nagyobb az influence */
				st_influence += 8;

				/** Lekéri az egyik elfordulás celláját */
				HexCell prev = location.getCell(direction.prev());
				/** Ha van ott cella, lekéri az influence-t */
				int p_influence = (prev == null) ? -2 : prev.getInfluence();
				/** Legalább 1 legyen az influence */
				p_influence += 1;

				/** Lekéri a másik fordulás celláját */
				HexCell next = location.getCell(direction.next());
				/** Ha van ott cella, lekéri az influence-t */
				int n_influence = (next == null) ? -2 : next.getInfluence();
				/** Legalább 1 legyen az influence */
				n_influence += 1;

				/**
				 * Ha engedélyezve van a nemdeterminisztikus működés, random
				 * szorzót kap
				 */
				if (Main.RND) {
					st_influence *= (int) (Math.random() * 10);
					p_influence *= (int) (Math.random() * 10);
					n_influence *= (int) (Math.random() * 10);
				}

				if (st_influence < 0 && p_influence < 0 && n_influence < 0) {
					/**
					 * Ha az összes érték érvénytelen, akkor csak forduljon
					 * egyet
					 */
					direction = direction.next();
				} else if (st_influence >= p_influence && st_influence > 0) {
					if (st_influence >= n_influence) {
						/** Az egyenes irány lett a kiválasztott */
						move(straight);
					} else if (n_influence > 0) {
						/**
						 * Ha kanyarodni kell, nem csak arra mozog, hanem az
						 * irányt is változtatja
						 */
						direction = direction.next();
						move(next);
					} else {
						/** Ha mégis érvénytelen lenne */
						direction = direction.next();
					}
				} else {
					if (p_influence >= n_influence && p_influence > 0) {
						/**
						 * Ha kanyarodni kell, nem csak arra mozog, hanem az
						 * irányt is változtatja
						 */
						direction = direction.prev();
						move(prev);
					} else if (n_influence > 0) {
						/**
						 * Ha kanyarodni kell, nem csak arra mozog, hanem az
						 * irányt is változtatja
						 */
						direction = direction.next();
						move(next);
					} else {
						/** Ha mégis érvénytelen lenne */
						direction = direction.next();
					}
				}
			} else if (state == AntState.AS_Returning) {
				/** Ha visszatérő módban mozog */

				/** Lekéri, hogy merre van a hangyaboly... */
				direction = location.getToAntHill();
				/** ...majd arra megy */
				move(location.getCell(direction));
			}
			/** Ha mozgott, akkor lenullázza a számlálóját */
			unusedTicks = 0;
		}
		Logger.exiting(this);
	}

	/**
	 * {@inheritDoc} Kiírja az irányát, ételét és sebességét is
	 * 
	 */
	public String toString() {
		return "ant " + direction + " " + (myFood == null ? 0 : 1) + " " + speed;
	}
}
