package main;

import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.io.InputStream;

import main.Visualizer.Images;
import main.Visualizer.Layer;
import main.Visualizer.MapCell;
import main.Visualizer.Role;

/**
 * Ez kezeli le külvilággal a kommunikációt.
 * 
 * @author Kozaroczy
 * 
 */
public class Controller extends MouseAdapter implements MouseListener {

	/** Az alapértelmezett szélesség, értéke {@value} */
	private static final int defaultX = 25;
	/** Az alapértelmezett magasság, értéke {@value} */
	private static final int defaultY = 13;
	/** Az időlépték */
	protected long timeStep = 150; // ms
	/** The singleton instance */
	private static Controller theOne;

	/** A játék fő állapotai */
	public enum GameState {
		GS_Before, GS_Running, GS_After
	}

	/** A játék aktuális állapota */
	GameState state;

	/**
	 * Szingleton példány elérése.
	 * 
	 * @return A szigleton példány.
	 */
	public static Controller getInstance() {
		if (theOne == null) {
			try {
				theOne = new Controller();
			} catch (SingletonException s) {
				System.err.println("Singleton error");
			}
		}
		return theOne;
	}

	/** A megjelenítő panel. */
	private Visualizer window;
	/** A többszörös hozzáférés kivédésére használt objektum */
	private Object initLock;

	/**
	 * Konstruktor.
	 * 
	 * @throws SingletonException
	 */
	public Controller() throws SingletonException {
		if (theOne != null)
			throw new SingletonException();
		theOne = this;
		window = new Visualizer();
		window.drawStartScreen();
		state = GameState.GS_Before;
		initLock = new Object();
		window.addMouseListener(this);
	}

	/**
	 * Kattintások kezelése
	 * 
	 * @param e
	 *            Az egérkattintás adatai
	 */
	public void mouseClicked(MouseEvent e) {
		/** Pozíciók lekérése */
		int x = e.getX();
		int y = e.getY();
		Role role = Role.NOTHING;
		switch (state) {
		/** A kezdőképernyő történései */
		case GS_Before:
			/** Megkeresi, hogy milyen szerepe van annak, amire kattintott a felhasználó */
			for (Images i : window.imagesToDraw.get(Layer.OVERLAY)) {
				if (i.contains(x, y)) {
					role = i.role;
					break;
				}
			}
			switch (role) {
			case PLAY:
				/** Indításra kattintott */
				startGame();
				window.drawField();
				state = GameState.GS_Running;
				break;
			case SETTINGS:
				/** Beállításokra kattintott */
				window.settingsMenu = !window.settingsMenu;
				break;
			case HIGHSCORE:
				/** Highscore-ra kattintott */
				// TODO: highscore menu
				break;
			case EXIT:
				/** Ki akar lépni. Nem kéne hagyni */
				if (window.confirmExit()) {
					System.exit(0);
				}
				break;
			case CHANGESPEED:
				/** Meg akarja változtatni a játék sebességét */
				timeStep = (timeStep + 50)%1000;
				timeStep = timeStep == 0 ? 250 : timeStep;
				break;
			default:
				/** Minden másra ne csináljon semmit */
				break;
			}
			break;

		/** A játék közbeni történések */
		case GS_Running:
			if (e.getButton() == MouseEvent.BUTTON1) {
				/** Ha bal gombbal kattint, fújunk a spray-el */
				HexCell c = null;
				/** Megkeressük, hogy melyik cellára kattintott */
				for (MapCell mc : window.cells) {
					if (mc.contains(x, y)) {
						c = mc.representedHexCell;
					}
				}
				if (c != null) {
					Director.getInstance().activeSpray.doSpraying(c);
				}
			} else if (e.getButton() == MouseEvent.BUTTON3) {
				/** Ha jobb gombbal, akkor spray-t váltunk */
				Director.getInstance().nextSpray();
			}
			break;

		/** A játék vége képernyő */
		case GS_After:
			/** Megkeressük, hogy mire kattintott */
			for (Images i : window.imagesToDraw.get(Layer.OVERLAY)) {
				if (i.contains(x, y)) {
					role = i.role;
					break;
				}
			}
			switch (role) {
			case PLAY:
				/** Új játék indítása. Boldogok vagyunk */
				state = GameState.GS_Before;
				startGame();
				window.drawField();
				state = GameState.GS_Running;
				break;
			case EXIT:
				/** Kilép. De biztos visszajön még */
				if (window.confirmExit()) {
					System.exit(0);
				}
				break;
			default:
				/** Egyébként ne csináljon semmit */
				break;
			}
			break;
		}
	}

	/**
	 * Ezen keresztül szól a Director, hogy elfogyott az étel a pályáról
	 */
	public void runOutOfFood() {
		state = GameState.GS_After;
	}

	/**
	 * Alias a startGame(null)-ra
	 * 
	 * @see #startGame(InputStream)
	 */
	public void startGame() {
		startGame(null);
	}

	/**
	 * Inicializálja a Directort és a visualizerPanelt.
	 * 
	 * @param commandInputStream
	 *            A betöltendő parancsok streamje. Ha null, akkor a default
	 *            pályagenerálás történik.
	 */
	public void startGame(InputStream commandInputStream) {
		Director director = Director.getInstance();

		/** Amíg új pályát rak össze, addig senki ne akarjon hozzányúlni */
		synchronized (initLock) {
			director.setSize(defaultX, defaultY);
			director.initialize();
		}

		/** Rajzoljuk újra a cellákat is, hogy meglegyenek az asszociációk */
		synchronized (window.cellsLock) {
			window.cells = new MapCell[0];
		}

		/** Legyen új spray, ebből az első az aktív */
		director.sprays.add(new AntScentVanisherSpray());
		director.sprays.add(new PoisonSpray());
		director.activeSpray = director.sprays.get(0);
	}

	/**
	 * A játék főciklusa.
	 */
	public void playGame() {
		/** Director, hogy ne kelljen mindig újra lekérdezni */
		Director director = Director.getInstance();

		/** Fusson a végtelenségig, máshol lépünk ki a programból */
		while (true) {
			long elapsed = System.currentTimeMillis();
			
			/** Ha fut, vagy a vége képernyő, akkor legyen tick() */
			if (state == GameState.GS_Running || state == GameState.GS_After) {
				Logger.begin();
				synchronized (initLock) {
					director.tick();
				}
				Logger.end();
			}
			
			/** Egyébként mindig rajzolja ki amit kell */
			switch (state) {
			case GS_Before:
				window.drawStartScreen();
				break;
			case GS_Running:
				window.drawField();
				break;
			case GS_After:
				synchronized (initLock) {
					window.drawEndScreen();
				}
				break;
			}
			
			/** Várakozzon annyit, hogy összességében timeStep múljon el */
			elapsed = System.currentTimeMillis() - elapsed;
			elapsed = elapsed < timeStep ? elapsed : timeStep;
			try {
				Thread.sleep(timeStep-elapsed);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
	}

}
