package main;

import java.util.LinkedList;

import main.HexCell.Direction;
import main.Visualizer.Layer;

/**
 * 
 * Hangyászsünöket reprezentáló osztály
 * 
 */
public class Echidna extends Wanderer {

	/**
	 * A hangyásszün lehetséges állapotai. Lehet ES_Wandering (Kóricáló) és
	 * ES_Leaving (távozó)
	 * 
	 */
	public enum EchidnaState {
		ES_Wandering, ES_Leaving
	}

	/** A megevett hangyák száma. Maximum {@value #maxAntsEaten}-t tud */
	protected int antsEaten;

	/** Maximum ennyi hangyát tud megenni, értéke {@value} */
	protected static final int maxAntsEaten = 8;
	/** Az alapértelmezett sebesség, értéke {@value} */
	protected static final int defaultSpeed = 1;

	/**
	 * A hangyászsün állapota
	 * 
	 * @see EchidnaState
	 */
	protected EchidnaState state;

	/** A kijutáshoz használt útvonal */
	protected LinkedList<HexCell> exitPath = null;

	/**
	 * Alapértelmezett konstruktor
	 * 
	 * @param location
	 *            A hangyászsün pozíciója
	 */
	Echidna(HexCell location) {
		super(location);
		Logger.entering(this, new Object[] { location });
		antsEaten = 0;
		state = EchidnaState.ES_Wandering;
		speed = defaultSpeed;
		Director.getInstance().incEchidna();
		Logger.exiting(this);
	}

	/**
	 * A hangyászsün betöltéshez használt konstruktora
	 * 
	 * @param location
	 *            Pozíció
	 * @param dir
	 *            Irány
	 * @param speed
	 *            Sebesség
	 * @param antsEaten
	 *            Megevett hangyák száma
	 */
	Echidna(HexCell location, Direction dir, int speed, int antsEaten) {
		super(location, dir, speed);
		Logger.entering(this, new Object[] { location, dir, speed, antsEaten });
		this.antsEaten = antsEaten;
		state = antsEaten == maxAntsEaten ? EchidnaState.ES_Leaving : EchidnaState.ES_Wandering;
		if (state == EchidnaState.ES_Wandering) {
			Director.getInstance().incEchidna();
		}
		Logger.exiting(this);
	}

	/**
	 * {@inheritDoc}
	 */
	public void accept(IInterActor interactor) {
		Logger.entering(this, new Object[] { interactor });
		interactor.interactWith(this);
		Logger.exiting(this);
	}

	/**
	 * {@inheritDoc}
	 */
	public String draw() {
		String suffixes = "_" + (antsEaten / 3) + "_" + direction.toString();
		return "img/echidna" + suffixes + ".png";
	}

	/**
	 * Megeszi a paraméterül kapott hangyát, majd ha szükséges módosítja az
	 * állapotát
	 * 
	 * @param ant
	 *            A megevésre szánt hangya
	 */
	public void eat(Ant ant) {
		Logger.entering(this, new Object[] { ant });
		ant.die();
		antsEaten++;
		if (antsEaten == maxAntsEaten) {
			state = EchidnaState.ES_Leaving;
			Director.getInstance().decEchidna();
		}
		Logger.exiting(this);
	}

	/**
	 * {@inheritDoc}
	 */
	public Layer getLayer() {
		return Layer.ACTOR;
	}

	/**
	 * {@inheritDoc} Ha még tud enni, akkor megeszi
	 */
	public void interactWith(Ant a) {
		Logger.entering(this, new Object[] { a });
		if (state == EchidnaState.ES_Wandering) {
			eat(a);
		}
		Logger.exiting(this);
	}

	/**
	 * {@inheritDoc} A sün meglöki a követ, ő maga ott marad, ha a kavics visszajönne majd az
	 * fogja visszábbtolni a hangyászsünt.
	 * 
	 * @see Pebble#interactWith(Echidna)
	 * @param p
	 *            A meglökendő kő
	 */
	public void interactWith(Pebble p) {
		Logger.entering(this, new Object[] { p });
		p.push(direction);
		Logger.exiting(this);
	}

	/**
	 * {@inheritDoc}
	 */
	public void interactWith(Snail s) {
		Logger.entering(this, new Object[] { s });
		move(location.getCell(direction.back()));
		Logger.exiting(this);
	}

	/**
	 * {@inheritDoc}
	 */
	public void move(HexCell target) {
		speed = defaultSpeed;
		super.move(target);
	}

	/**
	 * {@inheritDoc} Kóricáláskor véletlenszerű irányba mozog, távozáskor
	 * viszont megpróbál mindig a pálya széle felé haladni
	 */
	public void tick() {
		Logger.entering(this, null);
		unusedTicks++;
		if (unusedTicks >= speed) {
			if (state == EchidnaState.ES_Wandering) {
				/** Amíg kóricál, addig véletlenszerű irányba megy */
				if (Main.RND) {
					int rnd = (int) (Math.random() * Direction.values().length);
					for (int j = 0; j < rnd; j++) {
						direction = direction.next();
					}
				}
				HexCell new_loc = location.getCell(direction);
				if (new_loc != null) {
					move(new_loc);
				} else {
					direction = direction.next();
				}
			} else if (state == EchidnaState.ES_Leaving) {
				/** Ha távozóban van, céltudatos a mozása */
				Director d = Director.getInstance();
				if (exitPath == null) {
					do {
						int cellID;
						do {
							cellID = (int) Math.floor(Math.random() * (d.size_x) * (d.size_y));
						} while (!d.isExitID(cellID));

						int x = cellID % d.size_x;
						int y = (cellID - x) / d.size_x;
						HexCell exit = d.getCell(x, y);
						exitPath = location.aStar(exit);
					} while (exitPath == null);
				}
				if (exitPath != null) {
					if (exitPath.isEmpty()) {
						super.dispose();
						return;
					} else {
						HexCell nextCell = exitPath.removeFirst();
						for (Direction dir : location.adjacent.keySet()) {
							if (nextCell == location.adjacent.get(dir)) {
								direction = dir;
								break;
							}
						}

						if (nextCell != null && nextCell != location) {
							move(nextCell);
						}
					}
				}
			}
			unusedTicks = 0;
		}
		Logger.exiting(this);
	}

	/**
	 * {@inheritDoc}
	 */
	public String toString() {
		return "echidna " + this.direction + " " + this.antsEaten + " " + this.speed;
	}
}
