package main;

import java.util.ArrayList;
import java.util.HashMap;

import main.HexCell.Direction;
import main.Visualizer.Layer;

/**
 * 
 * Ételek szagát reprezentálja
 * 
 * @author Kozaroczy
 * 
 */
public class FoodScent extends Scent {

	private class Origin {
		protected Food originFood;
		protected HexCell originFoodCell;

		public Origin(Food originFood) {
			this.originFood = originFood;
			this.originFoodCell = originFood.location;
		}
	}

	/** A kibocsátó ételek */
	ArrayList<Origin> origins;

	/**
	 * Hangyák befolyásolásának mértéke.
	 * 
	 * @see Ant#move(HexCell)
	 */
	private static int influence = 50;

	/**
	 * Alapértelmezett konstruktor
	 * 
	 * @param location
	 *            Az új hely
	 * @param origin
	 *            Az szagot árasztó étel
	 * @param size
	 *            A terjedés mértéke
	 * @param remaining
	 *            A maradék idő
	 */
	public FoodScent(HexCell location, Food origin, int size, int remaining) {
		super(location, size, remaining);
		origins = new ArrayList<Origin>();
		addOrigin(origin);
	}

	/**
	 * Konstruktor terjedéshez
	 * 
	 * @param location
	 *            A pozíciója
	 * @param origin
	 *            Az eredeti kaja
	 * @param size
	 *            A terjedés mérete
	 * @param remaining
	 *            A maradék idő
	 * @param dir
	 *            A terjedés iránya
	 */
	public FoodScent(HexCell location, Food origin, int size, int remaining, Direction dir) {
		super(location, size, remaining, dir);
		Logger.entering(this, new Object[] { location, size, origin, dir });
		origins = new ArrayList<Origin>();
		addOrigin(origin);
		Logger.exiting(this);
	}

	/**
	 * {@inheritDoc}
	 */
	public void accept(IInterActor interactor) {
		Logger.entering(this, new Object[] { interactor });
		interactor.interactWith(this);
		Logger.exiting(this);
	}

	/**
	 * Új ételeket listaként is hozzá lehet adni
	 * @param originFood Lista az új szagforrásokról
	 */
	public void addOrigin(ArrayList<Origin> originFood) {
		Logger.entering(this, new Object[] { originFood });
		for (Origin o : originFood) {
			if (o.originFood.location == o.originFoodCell) {
				addOrigin(o.originFood);
			}
		}
		Logger.exiting(this);
	}

	/**
	 * Új ételt ad hozzá a forrásokhoz. Ezáltal összevonhatóak a szagok
	 * @param originFood Az új szagforrás
	 */
	public void addOrigin(Food originFood) {
		Logger.entering(this, new Object[] { originFood });
		boolean contains = false;
		for (Origin o : origins) {
			if (o.originFood == originFood) {
				contains = true;
			}
		}
		if (!contains) {
			origins.add(new Origin(originFood));
		}
		Logger.exiting(this);
	}

	/**
	 * {@inheritDoc} Kiüríti az ételek listáját is
	 */
	public void dispose() {
		origins.clear();
		super.dispose();
	}

	/**
	 * {@inheritDoc}
	 */
	public String draw() {
		return "img/foodscent.png";
	}

	/**
	 * {@inheritDoc}
	 */
	public Layer getLayer() {
		return Layer.SCENT;
	}

	/**
	 * A hangyák befolyásolása. Függ a befolyásoló képességtől és a szag
	 * intenzitásától
	 * 
	 * @see #influence
	 * @see Scent#remaining
	 */
	public int influence() {
		Logger.entering(this, null);
		Logger.exiting(this, influence * remaining * origins.size());
		return influence * remaining * origins.size();
	}

	/**
	 * Amíg az {@link #origins}-ből legalább egy a helyén van, nem csökken a
	 * szag intenzitása és terjeng az illat
	 * 
	 */
	public void tick() {
		Logger.entering(this, null);
		/** Ha már elfogyott, eltűnik */
		if (remaining <= 0) {
			dispose();
			Logger.exiting(this);
			return;
		}

		/** Ha valamelyik kibocsátó étel eltűnik, vegyük ki a listából */
		for (int i = origins.size() - 1; i >= 0; i--) {
			if (origins.get(i).originFood.location != origins.get(i).originFoodCell) {
				origins.remove(i);
			}
		}

		if (!origins.isEmpty()) {
			/** Ha van még kibocsátó... */
			if (radius > 0) {
				/** ... és tud terjedni */
				HashMap<Direction, HexCell> adjacent = location.getAdjacents();
				if (first) {
					/** Ha ő az első, minden irányba terjed */
					for (Direction d : Direction.values()) {
						HexCell c = adjacent.get(d);
						if (c != null) {
							new FoodScent(c, origins.get(0).originFood, radius - 1, remaining, d);
						}
					}
				} else {
					/** Egyébként csak a megfelelő irányokba terjed */
					for (int i = 0; i < dir.size(); i++) {
						HexCell c = adjacent.get(dir.get(i));
						if (c != null) {
							new FoodScent(c, origins.get(0).originFood, radius - 1, remaining, dir.get(i));
						}
					}
				}
				/** Ha terjedt, akkor többet ne terjedjen */
				radius = 0;
			}
		} else {
			/** Egyébként ne terjedjen, sőt kezdjen elillanni */
			remaining--;
		}
		Logger.exiting(this);
	}

	/**
	 * {@inheritDoc}
	 */
	public String toString() {
		return "foodscent " + radius + " " + remaining;
	}
}
