package main;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.PriorityQueue;

/**
 * 
 * A pálya celláit reprezentáló osztály
 * 
 */
public class HexCell {

	/**
	 * Az irányokat reprezentáló felsorolt típus. Rendelkezik csodaszép
	 * tagfügvényekkel is.
	 * 
	 * @see #back()
	 * @see #next()
	 * @see #prev()
	 */
	public enum Direction {
		NORTH, NORTH_EAST, SOUTH_EAST, SOUTH, SOUTH_WEST, NORTH_WEST;

		/**
		 * Stringből dönti el az értékét
		 * 
		 * @param s
		 *            A visszafejtendő string
		 */
		public static Direction fromString(String s) {
			switch (s) {
			case "n":
				return NORTH;
			case "ne":
				return NORTH_EAST;
			case "nw":
				return NORTH_WEST;
			case "s":
				return SOUTH;
			case "sw":
				return SOUTH_WEST;
			case "se":
				return SOUTH_EAST;
			default:
				return NORTH;
			}
		}

		/** Az ellenkező irányt adja vissza. */
		public Direction back() {
			Logger.entering(this, null);
			Direction ret = values()[(ordinal() + values().length / 2) % values().length];
			Logger.exiting(this, ret);
			return ret;
		}

		/** A következő irányt adja vissza. */
		public Direction next() {
			Logger.entering(this, null);
			Direction ret = values()[(ordinal() + 1) % values().length];
			Logger.exiting(this, ret);
			return ret;
		}

		/** Az előző irányt adja vissza. */
		public Direction prev() {
			Logger.entering(this, null);
			Direction ret = values()[(ordinal() + values().length - 1) % values().length];
			Logger.exiting(this, ret);
			return ret;
		}

		/** Szöveggé konvertálja az irányt */
		public String toString() {
			switch (this) {
			case NORTH:
				return "n";
			case NORTH_EAST:
				return "ne";
			case NORTH_WEST:
				return "nw";
			case SOUTH:
				return "s";
			case SOUTH_EAST:
				return "se";
			case SOUTH_WEST:
				return "sw";
			default:
				return "n";
			}
		}
	}

	/** A szomszédok */
	protected HashMap<Direction, HexCell> adjacent;
	/** A cellán tartózkodók listája */
	protected ArrayList<Located> contents;
	/** Igaz, ha ez a cella a pálya széle */
	protected boolean exit;
	/** A legközelebbi hangyaboly felé vezető útvonal */
	protected LinkedList<HexCell> pathToAntHill;
	/** A hangyabolyhoz vezető útvonal iránya */
	protected Direction directionToAntHill = Direction.NORTH;
	/** A cella pozíciója */
	protected int x, y;

	/** AStar függvény timeout-ja. */
	protected static long timeOut = 1000;

	/** Konstruktor */
	public HexCell() {
		Logger.entering(this, null);
		adjacent = new HashMap<Direction, HexCell>();
		contents = new ArrayList<Located>();
		x = y = 0;
		Logger.exiting(this);
	}

	/** Konstruktor */
	public HexCell(int x, int y) {
		Logger.entering(this, null);
		adjacent = new HashMap<Direction, HexCell>();
		contents = new ArrayList<Located>();
		this.x = x;
		this.y = y;
		Logger.exiting(this);
	}

	/**
	 * Szomszédos mező hozzáadása
	 * 
	 * @param direction
	 *            Az az irány, ahova hozzáadjuk szomszédnak.
	 * @param hexCell
	 *            Hozzáadandó szomszéd cell.
	 */
	public void addCell(Direction direction, HexCell hexCell) {
		Logger.entering(this, new Object[] { direction, hexCell });
		adjacent.put(direction, hexCell);
		Logger.exiting(this);
	}

	/**
	 * Új, pozícióval rendelkező {@link Located} objektum hozzáadása.
	 * 
	 * @param newItem
	 *            A hozzáadandó objectum.
	 */
	public void addObject(Located newItem) {
		Logger.entering(this, new Object[] { newItem });
		contents.add(newItem);
		Logger.exiting(this);
	}

	/**
	 * Súlyozott A* útkeresés saját cellából. A talált útvonal <strong>nem
	 * feltétlenül</strong> optimális. Timeout: Ha túl sok ideig fut, akkor
	 * nullal tér vissza.
	 * 
	 * @param target
	 *            A célcella.
	 * @return Láncolt lista, ami az útvonal összes celláját tartalmazza, vagy
	 *         null ha nincs útvonal. Ha időtúllépés történt, akkor null.
	 */
	public LinkedList<HexCell> aStar(final HexCell target) {
		Logger.entering(this, new Object[] { target });
		int dx = target.x - x;
		if (dx < 0)
			dx = -dx;
		int dy = target.y - y;
		if (dy < 0)
			dy = -dy;

		final double N = (dx + dy) * (2.0 / 3.0);
		class Node implements Comparable<Node> {
			HexCell cell;
			int depth;

			double f;
			Node parent;

			Node(int x, int y, HexCell cell, int depth, Node parent) {
				this.cell = cell;
				this.depth = depth;

				int dx = target.x - x;
				if (dx < 0)
					dx = -dx;
				int dy = target.y - y;
				if (dy < 0)
					dy = -dy;

				double w = (1 - (depth / N));
				if (w < 0)
					w = 0;

				this.f = depth + (1 + w) * (dx + dy) * (2.0 / 3.0); // f(x) =
																	// g(x) + (1
																	// + e*w) *
																	// h(x)

				this.parent = parent;
			}

			@Override
			public int compareTo(Node other) {
				if (this.f < other.f)
					return -1;
				if (this.f > other.f)
					return 1;
				return 0;
			}
		}

		PriorityQueue<Node> remaining = new PriorityQueue<Node>();
		HashMap<HexCell, Integer> proc = new HashMap<HexCell, Integer>();
		remaining.add(new Node(this.x, this.y, this, 0, null));
		Node goal = null;

		long startTime = System.currentTimeMillis();
		boolean tooLong = false;

		while (remaining.size() > 0) {

			// timeout
			tooLong = (System.currentTimeMillis() - startTime) > timeOut;
			if (tooLong) {
				Logger.meanwhiling("HexCell.aStar(HexCell) timeOut. Null is returned");
				System.err.println("HexCell.aStar(HexCell) timeOut. Null is returned");
				Logger.exiting(this, null);
				return null;
			}

			Node current = remaining.remove();
			if (current.cell == target) {
				goal = current;
				break;
			}
			if (proc.containsKey(current.cell) && proc.get(current.cell) < current.depth)
				continue;
			proc.put(current.cell, current.depth);
			for (Direction d : Direction.values()) {
				HexCell next = current.cell.getCell(d);
				if (next != null && (!proc.containsKey(next) || (proc.containsKey(next) && proc.get(next) > current.depth + 1))) {
					boolean blocked = false;
					for (Located l : next.contents) {
						if (l instanceof Barrier) {
							blocked = true;
							break;
						}
					}
					if (!blocked)
						remaining.add(new Node(next.x, next.y, next, current.depth + 1, current));
					else
						proc.put(next, 0);
				}
			}
		}
		if (goal == null) {
			Logger.exiting(this, null);
			return null;
		}
		LinkedList<HexCell> result = new LinkedList<HexCell>();
		while (goal != null) {
			result.addFirst(goal.cell);
			goal = goal.parent;
		}
		Logger.exiting(this, result);
		return result;
	}

	/**
	 * Visszaadja az összes szomszédját.
	 * 
	 * @return A cella összes szomszédját tartalmazó, Direction-nel indexelt
	 *         HashMap.
	 */
	public HashMap<Direction, HexCell> getAdjacents() {
		Logger.entering(this, null);
		Logger.exiting(this, adjacent);
		return adjacent;
	}

	/**
	 * Visszaadja az összes elemet a {@link #contents} listából.
	 * 
	 * @return A cellán tartózkodó összes elem
	 * 
	 */
	public ArrayList<Located> getAll() {
		Logger.entering(this, null);
		Logger.exiting(this, contents);
		return contents;
	}

	/**
	 * Az adott irányban lévő cellával tér vissza.
	 * 
	 * @param dir
	 *            A keresett irány
	 * @return Az adott irányban lévő szomszédja
	 */
	public HexCell getCell(Direction dir) {
		Logger.entering(this, new Object[] { dir });
		Logger.exiting(this, adjacent.get(dir));
		return adjacent.get(dir);
	}

	/**
	 * Visszaadja, hogy az adott mező mennyire befolyásolja a hangya irányát.
	 * 
	 * @return A befolyásolás mértéke
	 * 
	 */
	public int getInfluence() {
		Logger.entering(this, null);
		int sum = 0;
		for (Located i : contents) {
			sum += i.influence();
		}
		Logger.exiting(this, sum);
		return sum;
	}

	/** Visszaadja, hogy milyen irányban van a hangyaboly */
	public Direction getToAntHill() {
		Logger.entering(this, null);
		Logger.exiting(this, directionToAntHill);
		return directionToAntHill;
	}

	/**
	 * X koordináta
	 * 
	 * @return x koordináta
	 */
	public int getX() {
		return x;
	}

	/**
	 * Y koordináta
	 * 
	 * @return y koordináta
	 */
	public int getY() {
		return y;
	}

	/**
	 * Ellenőrzés szempontjából lekérhető, hogy tényleg szomszédos-e a két
	 * cella.
	 * 
	 * @param other
	 *            A másik cella
	 */
	public boolean isAdjacent(HexCell other) {
		Logger.entering(this, new Object[] { other });
		boolean ret = adjacent.containsValue(other);
		Logger.exiting(this, ret);
		return ret;
	}

	/** Igazzal tér vissza, ha az adott cella kijárat. */
	public boolean isExit() {
		Logger.entering(this, null);
		Logger.exiting(this, exit);
		return exit;
	}

	/**
	 * Az adott elemet eltávolítja a {@link #contents} listáról.
	 * 
	 * @param toRemove
	 *            Az eltávolítandó Located referenciája.
	 */
	public void removeObject(Located toRemove) {
		Logger.entering(this, new Object[] { toRemove });
		contents.remove(toRemove);
		toRemove.location = null;
		Logger.exiting(this);
	}

	/**
	 * Beállítja, hogy az adott cella kijárat-e.
	 * 
	 * @param newVal
	 *            Igaz - A cella kijárat, Hamis - A cella nem kijárat.
	 */
	public void setExit(boolean newVal) {
		Logger.entering(this, new Object[] { newVal });
		exit = newVal;
		Logger.exiting(this);
	}

	/**
	 * Hangyabolyba vezető útvonal eltárolása
	 * 
	 * @param path
	 *            Az útvonal.
	 * @param direction
	 *            A következő cella iránya.
	 */
	@SuppressWarnings("unchecked")
	public void setPath(LinkedList<HexCell> path, Direction direction) {
		pathToAntHill = (LinkedList<HexCell>) path.clone();
		directionToAntHill = direction;
	}

}
