package main;

import java.awt.Color;
import java.awt.Cursor;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Polygon;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.EnumMap;
import java.util.HashMap;

import javax.imageio.ImageIO;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JPanel;

import main.Controller.GameState;

/**
 * A játék megjelenéséért felel
 * 
 * @author Kozaroczy
 * 
 */
@SuppressWarnings("serial")
public class Visualizer extends JPanel {
	private class DoubleColor {
		double r, g, b;

		DoubleColor(double r, double g, double b) {
			this.r = r;
			this.g = g;
			this.b = b;
		}

		DoubleColor m(DoubleColor o) {
			this.r = r * o.r;
			this.g = g * o.g;
			this.b = b * o.b;
			return this;
		}

		int toIntColor() {
			int ri = r < 0 ? 0 : (r > 1 ? 255 : (int) (r * 255));
			int gi = g < 0 ? 0 : (g > 1 ? 255 : (int) (g * 255));
			int bi = b < 0 ? 0 : (b > 1 ? 255 : (int) (b * 255));
			return (ri << 16) + (gi << 8) + bi;
		}
	}

	/**
	 * Minden kép kap egy szerepet, amit be kell töltenie. Innen tudja a
	 * Controller, hogy mit kell csinálnia.
	 * 
	 * @author Kozaroczy
	 * 
	 */
	public enum Role {
		PLAY, SETTINGS, HIGHSCORE, EXIT, CELL, CHANGESPEED, NOTHING;
	}

	/**
	 * A kirajzolást rétegenként végzi
	 */
	public enum Layer {
		BACKGROUND, SCENT, OBSTACLE, ACTOR, OVERLAY;
	}

	/**
	 * Egy kirajzolandó kép csomagoló osztálya.
	 */
	public class Images {
		int x, y;
		String imageString;
		HexCell position;
		Role role;

		public Images(int x, int y, String img, HexCell pos) {
			this.x = x;
			this.y = y;
			this.imageString = img;
			this.position = pos;
			role = Role.CELL;
		}

		public Images(int x, int y, String img, HexCell pos, Role function) {
			this.x = x;
			this.y = y;
			this.imageString = img;
			this.position = pos;
			role = function;
		}

		public void draw(Graphics page) {
			BufferedImage img = null;
			if (bufferedImages.containsKey(imageString)) {
				img = bufferedImages.get(imageString);
			} else {
				try {
					img = ImageIO.read(new File(imageString));
					bufferedImages.put(imageString, img);
				} catch (IOException e) {
					System.err.println("Problem with loading " + imageString);
				}
			}
			page.drawImage(img, x, y, frame);
		}

		public boolean contains(int x, int y) {
			if (this.x <= x && this.y <= y) {
				int w, h;
				h = bufferedImages.get(imageString).getHeight();
				w = bufferedImages.get(imageString).getWidth();
				return (this.x + w >= x) && (this.y + h >= y);
			} else {
				return false;
			}
		}
	}

	/**
	 * Egy HexCell képe.
	 */
	public class MapCell {
		public Polygon cell;
		public Color background;
		public boolean isFilled;
		public HexCell representedHexCell;

		MapCell(Polygon polyCell, HexCell representedHC) {
			this.cell = polyCell;
			this.isFilled = false;
			this.representedHexCell = representedHC;
		}

		public void draw(Graphics page) {
			page.setColor(Color.BLACK);
			page.drawPolygon(cell);
			if (isFilled) {
				page.setColor(this.background);
				page.fillPolygon(cell);
			}
		}

		public boolean contains(int x, int y) {
			return cell.contains(x, y);
		}

	}

	/**
	 * Perlin zaj a procedurális háttérgeneráláshoz.
	 * 
	 * @author Baliame
	 */
	private class NoiseGenerator {
		int i, j, k;
		int seed;
		double u, v, w, s;
		double onethird, onesixth;
		int[] T = new int[8];
		int[] A = new int[3];

		NoiseGenerator(int seed) {
			A[0] = A[1] = A[2] = 0;
			onethird = 1 / 3.0f;
			onesixth = 1 / 6.0f;
			T[0] = 0x15;
			T[1] = 0x38;
			T[2] = 0x32;
			T[3] = 0x2c;
			T[4] = 0x0d;
			T[5] = 0x13;
			T[6] = 0x07;
			T[7] = 0x2a;
			this.seed = seed;
		}

		int b(int N, int B) {
			return N >> B & 1;
		}

		int b(int i, int j, int k, int B) {
			return T[b(i, B) << 2 | b(j, B) << 1 | b(k, B)];
		}

		int fastfloor(double n) {
			return n > 0 ? (int) n : (int) n - 1;
		}

		double fk(int a) {
			s = (A[0] + A[1] + A[2]) * onesixth;
			double x = u - A[0] + s;
			double y = v - A[1] + s;
			double z = w - A[2] + s;
			double t = 0.6f - x * x - y * y - z * z;
			int h = shuffle(i + A[0], j + A[1], k + A[2]);
			A[a]++;
			if (t < 0)
				return 0;
			int b5 = h >> 5 & 1;
			int b4 = h >> 4 & 1;
			int b3 = h >> 3 & 1;
			int b2 = h >> 2 & 1;
			int b = h & 3;
			double p = b == 1 ? x : b == 2 ? y : z;
			double q = b == 1 ? y : b == 2 ? z : x;
			double r = b == 1 ? z : b == 2 ? x : y;
			p = b5 == b3 ? -p : p;
			q = b5 == b4 ? -q : q;
			r = b5 != (b4 ^ b3) ? -r : r;
			t *= t;
			return 8 * t * t * (p + (b == 0 ? q + r : b2 == 0 ? q : r));
		}

		double noise(double x, double y) {
			double z = seed;
			s = (x + y + z) * onethird;
			i = fastfloor(x + s);
			j = fastfloor(y + s);
			k = fastfloor(z + s);

			s = (i + j + k) * onesixth;
			u = x - i + s;
			v = y - j + s;
			w = z - k + s;

			A[0] = A[1] = A[2] = 0;

			int hi = u >= w ? u >= v ? 0 : 1 : v >= w ? 1 : 2;
			int lo = u < w ? u < v ? 0 : 1 : v < w ? 1 : 2;

			return fk(hi) + fk(3 - hi - lo) + fk(lo) + fk(0);
		}

		int shuffle(int i, int j, int k) {
			return b(i, j, k, 0) + b(j, k, i, 1) + b(k, i, j, 2) + b(i, j, k, 3) + b(j, k, i, 4) + b(k, i, j, 5) + b(i, j, k, 6) + b(j, k, i, 7);
		}
	}

	/** A pálya mérete */
	private int w, h;
	/** A megjelenítés ablaka */
	private JFrame frame;
	/** Kirajzolandó képek */
	protected EnumMap<Layer, ArrayList<Images>> imagesToDraw;
	/** A betöltött képek a nevük szerint hashMap-ben letárolva */
	private HashMap<String, BufferedImage> bufferedImages;
	/** Háttér külön letárolva */
	private BufferedImage backgroundImage;
	/** A hatszögháló */
	public MapCell[] cells;
	/** Lock a hatszöghálóhoz */
	protected Object cellsLock;
	/** Setting menü megjelenítése */
	protected boolean settingsMenu = false;

	/**
	 * Konstruktor
	 */
	public Visualizer() {
		cellsLock = new Object();
		frame = new JFrame();
		frame.add(this);
		frame.setVisible(false);
		frame.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
		frame.setResizable(false);
		frame.addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing(WindowEvent arg0) {
				if (confirmExit()) {
					System.exit(0);
				}
			}
		});

		setCursor(new Cursor(Cursor.CROSSHAIR_CURSOR));
		imagesToDraw = new EnumMap<Layer, ArrayList<Images>>(Layer.class);
		for (Layer l : Layer.values()) {
			imagesToDraw.put(l, new ArrayList<Images>());
		}
		bufferedImages = new HashMap<String, BufferedImage>();
		cells = new MapCell[0];
	}

	/**
	 * A játéktér kirajzolása
	 */
	public void drawField() {
		/** Színek beállítása */
		DoubleColor grass = new DoubleColor(0, 1.2, 0);
		DoubleColor dirt = new DoubleColor(1.22, 0.92, 0.66);

		/** Méretek beállítása */
		boolean generateBackground = (w == 0 || h == 0);
		w = Director.getInstance().size_x;
		h = Director.getInstance().size_y;
		int pw = w * 45 + 90;
		int ph = h * 52 + 120;

		/** Háttér generálása */
		if (generateBackground) {
			backgroundImage = new BufferedImage(pw, ph, BufferedImage.TYPE_INT_RGB);
			NoiseGenerator noiseGrass = new NoiseGenerator((int) (Math.random() * 100));
			NoiseGenerator noiseDirt = new NoiseGenerator((int) (Math.random() * 100));
			double tsw = Math.random() * 140 + 100;
			double tsh = Math.random() * 140 + 100;
			for (int y = 0; y < ph; y++) {
				for (int x = 0; x < pw; x++) {
					double n = noiseGrass.noise(x / tsw, y / tsh);
					DoubleColor c;
					if (n < -0.01) {
						double d = (noiseDirt.noise(x, y) + 1) / 2;
						c = new DoubleColor(d, d, d).m(dirt);
					} else if (n > 0.01) {
						double g = (noiseGrass.noise(x, y) + 1) / 2;
						c = new DoubleColor(g, g, g).m(grass);
					} else {
						double b = Math.random();
						if (b < 0.5) {
							double d = (noiseDirt.noise(x, y) + 1) / 2;
							c = new DoubleColor(d, d, d).m(dirt);
						} else {
							double g = (noiseGrass.noise(x, y) + 1) / 2;
							c = new DoubleColor(g, g, g).m(grass);
						}
					}
					int color = c.toIntColor();
					backgroundImage.setRGB(x, y, color);
				}
			}
		}

		/** Rácsok kiszámolása */
		synchronized (cellsLock) {
			if (cells == null || cells.length != w * h) {
				cells = new MapCell[w * h];
				for (int cy = 1; cy <= h; cy++) {
					for (int cx = 1; cx <= w; cx++) {
						int x = cx - 1;
						int y = cy - 1;
						if (x % 2 == 0) {
							cells[x + y * w] = new MapCell(new Polygon(new int[] { x * 45 + 30, x * 45 + 45, x * 45 + 75, x * 45 + 90, x * 45 + 75, x * 45 + 45 }, new int[] { y * 52 + 56, y * 52 + 30, y * 52 + 30, y * 52 + 56, y * 52 + 82, y * 52 + 82 }, 6), Director.getInstance().getCell(x, y));
						} else {
							cells[x + y * w] = new MapCell(new Polygon(new int[] { x * 45 + 30, x * 45 + 45, x * 45 + 75, x * 45 + 90, x * 45 + 75, x * 45 + 45 }, new int[] { y * 52 + 82, y * 52 + 56, y * 52 + 56, y * 52 + 82, y * 52 + 108, y * 52 + 108 }, 6), Director.getInstance().getCell(x, y));
						}
					}
				}
			}
		}

		/** Cellák feldolgozása */
		synchronized (imagesToDraw) {
			/** Előző rajzok törlése */
			for (Layer l : Layer.values())
				imagesToDraw.get(l).clear();

			/** Feltöltés új rajzokkal */
			for (int i = 0; i < w; i++) {
				for (int j = 0; j < h; j++) {
					HexCell c = Director.getInstance().getCell(i, j);
					for (Located l : c.getAll()) {
						int locx;
						int locy;
						if (i % 2 == 0) {
							locx = i * 45 + 44;
							locy = j * 52 + 40;
						} else {
							locx = i * 45 + 44;
							locy = j * 52 + 66;
						}
						imagesToDraw.get(l.getLayer()).add(new Images(locx, locy, l.draw(), c));
					}
				}
			}
		}

		/** Maradékok kirajzolása */
		int x = 0;
		boolean firstActive = true;
		synchronized (imagesToDraw) {
			for (int i = 0; i < Director.getInstance().sprays.size(); i++) {
				Spray current = Director.getInstance().sprays.get(i);
				for (int j = 0; j < current.remaining; j++) {
					if (firstActive && current == Director.getInstance().activeSpray) {
						firstActive = false;
						imagesToDraw.get(Layer.OVERLAY).add(new Images(x, 0, "img/shadow.png", null, Role.NOTHING));
					}
					imagesToDraw.get(Layer.OVERLAY).add(new Images(x, 0, Director.getInstance().sprays.get(i).draw(), null, Role.NOTHING));
					x += 30;
				}
			}
		}

		/** Láthatóvá tétel */
		frame.setSize(pw, ph);
		setSize(pw, ph);
		frame.setVisible(true);
	}

	/**
	 * A kezdőképernyő kirajzolása
	 */
	public void drawStartScreen() {
		/** Méretek beállítása */
		frame.setSize(1000, 550);
		setSize(1000, 550);

		/** Megjelenítés */
		synchronized (imagesToDraw) {
			/** Előző rajzok törlése */
			for (Layer l : Layer.values())
				imagesToDraw.get(l).clear();

			/** Új rajzok hozzáadása */
			imagesToDraw.get(Layer.OVERLAY).add(new Images(380, 96, "img/play.png", null, Role.PLAY));
			imagesToDraw.get(Layer.OVERLAY).add(new Images(380, 192, "img/settings.png", null, Role.SETTINGS));
			imagesToDraw.get(Layer.OVERLAY).add(new Images(380, 288, "img/highscore.png", null, Role.HIGHSCORE));
			imagesToDraw.get(Layer.OVERLAY).add(new Images(380, 384, "img/exit.png", null, Role.EXIT));
			imagesToDraw.get(Layer.BACKGROUND).add(new Images(0, 0, "img/startscreen.png", null, Role.NOTHING));
			
			if(settingsMenu){
				imagesToDraw.get(Layer.OVERLAY).add(new Images(120, 192, "img/empty.png", null, Role.CHANGESPEED));
			}
		}

		/** Láthatóvá tétel */
		frame.setVisible(true);
	}

	/**
	 * A képek bezárása
	 */
	public void finalize() {
		ArrayList<BufferedImage> imgsToClose = (ArrayList<BufferedImage>) bufferedImages.values();
		while (imgsToClose.size() > 0) {
			imgsToClose.remove(0);
		}
	}

	/**
	 * Tényleges kirajzolás
	 */
	public void paintComponent(Graphics page) {
		super.paintComponent(page);
		page.drawImage(backgroundImage, 0, 0, w * 45 + 90, h * 52 + 120, this);
		page.setColor(Color.BLACK);

		/** Hatszögháló kirajzolása */
		synchronized (cellsLock) {
			for (int y = 0; y < h; y++) {
				for (int x = 0; x < w; x++) {
					cells[x + y * w].draw(page);
				}
			}
		}

		page.setFont(new Font("alma",Font.BOLD,40));
		
		/** A kirajzolandó képek kirajzolása */
		synchronized (imagesToDraw) {
			for (Layer l : Layer.values()) {
				for (Images i : imagesToDraw.get(l)) {
					i.draw(page);
					if(settingsMenu && i.role == Role.CHANGESPEED){
						Integer out = new Integer((int) Controller.getInstance().timeStep);
						page.drawString(out.toString(), i.x + 90, i.y + 40);
					}
				}
			}
		}
		
		/** Ételek számának kirajzolása */
		if(Controller.getInstance().state == GameState.GS_Running){
			Integer out = new Integer((int) Director.getInstance().foodRemaining);
			page.drawString(": " +out.toString(),(w-1) * 45, 40);
			new Images((w-2) * 45,0,"img/food00.png",null).draw(page);
		}
	}

	/**
	 * Játék vége képernyő kirajzolása, a háttérben tovább fut a játék
	 */
	public void drawEndScreen() {
		/** Színek beállítása */
		DoubleColor grass = new DoubleColor(0, 1.2, 0);
		DoubleColor dirt = new DoubleColor(1.22, 0.92, 0.66);

		/** Méretek beállítása */
		boolean generateBackground = (w == 0 || h == 0);
		w = Director.getInstance().size_x;
		h = Director.getInstance().size_y;
		int pw = w * 45 + 90;
		int ph = h * 52 + 120;

		/** Háttér generálása */
		if (generateBackground) {
			backgroundImage = new BufferedImage(pw, ph, BufferedImage.TYPE_INT_RGB);
			NoiseGenerator noiseGrass = new NoiseGenerator((int) (Math.random() * 100));
			NoiseGenerator noiseDirt = new NoiseGenerator((int) (Math.random() * 100));
			double tsw = Math.random() * 140 + 100;
			double tsh = Math.random() * 140 + 100;
			for (int y = 0; y < ph; y++) {
				for (int x = 0; x < pw; x++) {
					double n = noiseGrass.noise(x / tsw, y / tsh);
					DoubleColor c;
					if (n < -0.01) {
						double d = (noiseDirt.noise(x, y) + 1) / 2;
						c = new DoubleColor(d, d, d).m(dirt);
					} else if (n > 0.01) {
						double g = (noiseGrass.noise(x, y) + 1) / 2;
						c = new DoubleColor(g, g, g).m(grass);
					} else {
						double b = Math.random();
						if (b < 0.5) {
							double d = (noiseDirt.noise(x, y) + 1) / 2;
							c = new DoubleColor(d, d, d).m(dirt);
						} else {
							double g = (noiseGrass.noise(x, y) + 1) / 2;
							c = new DoubleColor(g, g, g).m(grass);
						}
					}
					int color = c.toIntColor();
					backgroundImage.setRGB(x, y, color);
				}
			}
		}

		/** Rácsok kiszámolása */
		synchronized (cellsLock) {
			if (cells == null || cells.length != w * h) {
				cells = new MapCell[w * h];
				for (int cy = 1; cy <= h; cy++) {
					for (int cx = 1; cx <= w; cx++) {
						int x = cx - 1;
						int y = cy - 1;
						if (x % 2 == 0) {
							cells[x + y * w] = new MapCell(new Polygon(new int[] { x * 45 + 30, x * 45 + 45, x * 45 + 75, x * 45 + 90, x * 45 + 75, x * 45 + 45 }, new int[] { y * 52 + 56, y * 52 + 30, y * 52 + 30, y * 52 + 56, y * 52 + 82, y * 52 + 82 }, 6), Director.getInstance().getCell(x, y));
						} else {
							cells[x + y * w] = new MapCell(new Polygon(new int[] { x * 45 + 30, x * 45 + 45, x * 45 + 75, x * 45 + 90, x * 45 + 75, x * 45 + 45 }, new int[] { y * 52 + 82, y * 52 + 56, y * 52 + 56, y * 52 + 82, y * 52 + 108, y * 52 + 108 }, 6), Director.getInstance().getCell(x, y));
						}
					}
				}
			}
		}

		/** Cellák feldolgozása */
		synchronized (imagesToDraw) {
			/** Előző rajzok törlése */
			for (Layer l : Layer.values())
				imagesToDraw.get(l).clear();

			/** Feltöltés új rajzokkal */
			for (int i = 0; i < w; i++) {
				for (int j = 0; j < h; j++) {
					HexCell c = Director.getInstance().getCell(i, j);
					for (Located l : c.getAll()) {
						int locx;
						int locy;
						if (i % 2 == 0) {
							locx = i * 45 + 44;
							locy = j * 52 + 40;
						} else {
							locx = i * 45 + 44;
							locy = j * 52 + 66;
						}
						imagesToDraw.get(l.getLayer()).add(new Images(locx, locy, l.draw(), c));

					}
				}
			}
		}

		/** Vezérlők az előtérbe */
		synchronized (imagesToDraw) {
			/** Új rajzok hozzáadása */
			imagesToDraw.get(Layer.OVERLAY).add(new Images(366, 96, "img/gameover.png", null, Role.NOTHING));
			imagesToDraw.get(Layer.OVERLAY).add(new Images(200, 384, "img/newgame.png", null, Role.PLAY));
			imagesToDraw.get(Layer.OVERLAY).add(new Images(774, 384, "img/exit.png", null, Role.EXIT));

		}

		/** Láthatóá tétel */
		frame.setSize(pw, ph);
		setSize(w * 60 + 60, h * 52 + 60);
		frame.setVisible(true);
	}

	/**
	 * A kilépést megerősítő ablak
	 * 
	 * @return Igaz, ha tényleg ki akar lépni
	 */
	public boolean confirmExit() {
		int confirm = JOptionPane.showOptionDialog(this, "Are you sure to close this exciting, super game?", "Exit Confirmation", JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE, null, null, null);
		if (confirm == JOptionPane.YES_OPTION) {
			return true;
		}
		return false;
	}

}
